/*-
 * Copyright (C)2003..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2003..2021 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: read.c,v 1.54 2021/06/12 00:36:47 babolo Exp $\n"

#define BLIN_COMPAT   4
#define MIFE_COMPAT   VMAJOR
#define MIFE_INTERNAL 1

#include <sys/resource.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <malloc_np.h>
#include <babolo/BLINflag.h>
#include "mife.h"

#define C_zr 0x02 /* empty buffer    */
#define C_al 0x04 /* allocate buffer */
#define C_mv 0x08 /* shift buffer    */
#define C_ra 0x10 /* realloc buffer  */

static ssize_t
rd(mife_descriptor *mife, void *buf, ssize_t len) {
#   define blin_internal_flags (mife->flags)
    int              retrylim;
    ssize_t          rdlen;
    struct timespec  tick;
    struct timespec *t;
    int              i;

    ifBLIN_QX2("+ %"BLIN_X" buf=%"BLIN_X" len=%"BLIN_D, BLIN_I(mife), BLIN_I(buf), len);
    retrylim = (MIFE_RTRY & mife->flags);
    do  { /* Read something only onse */
        /*******************************************************************   */
        if  (!mife->offset && !mife->actlen) {
            t = &mife->itime;
        } else {
            t = &mife->otime;
        }
        if  (t->tv_sec || t->tv_nsec) {
            if  (0 > clock_gettime(CLOCK_MONOTONIC_FAST, &tick)) {
                ifBLIN_QW1("clock_gettime #1");
            }
            ifBLIN_QX5("  kq=%d @ %u.%09lu", mife->kev->kq, tick.tv_sec, tick.tv_nsec);
            EV_SET(&mife->kev->kev, mife->f.fd, EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0, NULL);
            i = kevent(mife->kev->kq, &mife->kev->kev, 1, &mife->kev->kev, 1, t);
            if  (0 > i) {
                ifBLIN_QW1("kevent");
                rdlen = -EX_IOERR;
                goto out;
            } else if (!i) {
                rdlen = 0;
                if  (0 > clock_gettime(CLOCK_MONOTONIC_FAST, &tick)) {
                    ifBLIN_QW1("clock_gettime #2");
                }
                ifBLIN_QX1("timeout @ %"BLIN_U".%09lu", tick.tv_sec, tick.tv_nsec);
                break;
            }
            if  (0 > clock_gettime(CLOCK_MONOTONIC_FAST, &tick)) {
                ifBLIN_QW1("clock_gettime #3");
            }
            ifBLIN_QX5( "  offset="BLIN_O"%d @ %"BLIN_U".%09lu"
                      , mife->offset
                      , tick.tv_sec
                      , tick.tv_nsec
                      );
        }/****************************************************************************************/
        rdlen = mife->f.read(mife->f.fd, buf, len);
        ifBLIN_QX2( "  -> read(%d, %"BLIN_X"[%"BLIN_D"])=%"BLIN_D" +%"BLIN_O"d retrylim=%d"
                      , mife->f.fd
                      , buf
                      , len
                      , rdlen
                      , mife->offset
                      , retrylim
                      )
            ;
        if  (rdlen < 0) {
            if  (errno == EAGAIN || errno == EDEADLK) {
                ifBLIN_QW1("read retry");
                rdlen = 0;
            } else {
                ifBLIN_QW0("read");
                rdlen = -EX_IOERR;
                goto out;
        }   }
        if (!rdlen) {
            if  (retrylim-- <= 0) break;
            usleep(1 << ((MIFE_RTRY & mife->flags) - (retrylim / 2)));
        }
        ifBLIN_QX3("rdlen=%"BLIN_D, rdlen);
    } while (!rdlen);
    if  (!rdlen) mife->flags |= MIFE_EOFL;
out:
    ifBLIN_QX2("- %"BLIN_D, rdlen);
    return(rdlen);
#   undef blin_internal_flags
}

static ssize_t
mife_pipe(mife_descriptor *mife, ssize_t ordlen, off_t offset) {
#   define blin_internal_flags (mife->flags)
    ssize_t          needlen;
    ssize_t          ex = -1;
    ssize_t          tmplen;
    ssize_t          trylen;
    void            *tmpbuf;
    ssize_t          rdlen;
    off_t            delta = 0;;
    size_t           plb;

    ifBLIN_QX2("+ %"BLIN_X" ordlen=%"BLIN_D" offset=%"BLIN_O"d", BLIN_I(mife), ordlen, offset);
    MIFE_DUMP(mife, delta);
    if  (mife->offset > offset) {
        ifBLIN_QX0("mife->offset > offset");
        ERROUT(-EX_USAGE, EINVAL);
    }
    if  (!ordlen) {
        needlen = (mife->flags & MIFE_SLID) ? mife->mindelta : 0;
    } else if (ordlen < mife->mindelta) {
        needlen = mife->mindelta;
    } else {
        needlen = ordlen;
    }
    plb = (mife->flags & MIFE_FULL) ? 1 : 0;
    for (;;) {
        trylen = (MAXPACKT * 2 - 1 + needlen) & ~(MAXPACKT - 1);
        tmplen = (MAXPACKT * 2 - 1 + needlen + plb) & ~(MAXPACKT - 1);
        if  (tmplen < mife->mindelta) tmplen = mife->mindelta;
        mife->ofpoint = offset;
        delta = mife->ofpoint - mife->offset;
        MIFE_DUMP(mife, delta);
        ifBLIN_QX2("needlen %"BLIN_D" tmplen %"BLIN_D" trylen %"BLIN_D, needlen, tmplen, trylen);
        if  (!mife->buffer || !mife->buflen || (delta >= (off_t)mife->actlen)) {
            if  (mife->eol) mife->eol->stop(mife);
            mife->offset += mife->actlen;
            delta = mife->ofpoint - mife->offset;
            mife->actbeg = 0;
            mife->actlen = 0;
        }
        if  (!mife->buffer || !mife->buflen) {
            if  (mife->flags & MIFE_ALLC) free(mife->buffer);
            mife->flags |= MIFE_ALLC;
            mife->buflen = trylen;
            if  (!(mife->buffer = malloc(tmplen))) {
                ifBLIN_QW0("malloc(%"BLIN_U")", mife->buflen);
                ERROUT(-EX_OSERR, ENOMEM);
        }   }
        if  (delta >= (off_t)mife->actlen) {
            while(delta >= (off_t)mife->actlen) {
                mife->offset += mife->actlen;
                delta = mife->ofpoint - mife->offset;
                if  (0 > (rdlen = rd(mife, mife->buffer, trylen))) {
                    ex = rdlen;
                    goto out;
                }
                mife->actlen = rdlen;
                MIFE_DUMP(mife, delta);
                if  (!rdlen) {
                    if  (mife->ofpoint > mife->offset) {
                        mife->ofpoint = mife->offset;
                        delta = 0;
                        ex = -EX_NOINPUT;
                        errno = ESPIPE;
                    } else {
                        ex = 0;
                        if  (mife->flags & MIFE_FULL) CHARBUF[mife->actbeg] = 0;
                    }
                    goto out;
            }   }
            if  (mife->flags & MIFE_FULL) CHARBUF[mife->actbeg + mife->actlen] = 0;
    ;       if  (mife->eol) mife->eol->start(mife, mife->actbeg + delta);
    ;       ifBLIN_QX3("needlen %"BLIN_D" tmplen %"BLIN_D" trylen %"BLIN_D, needlen, tmplen, trylen);
    ;       if  (mife->eol && (0 <= mife->eol->detect(mife))) {
    ;           ifBLIN_QX5(" eolp=%"BLIN_D", eolst=%"BLIN_D, mife->eol->p, mife->eol->st);
                ex = mife->actlen - delta;
                goto out;
            }
    ;       if  (ordlen && (((off_t)mife->actlen - delta) >= ordlen)) {
                ex = mife->actlen - delta;
                goto out;
        }   }
        if  (((ssize_t)mife->actbeg + (ssize_t)delta) > mife->mindelta) {
            bcopy(mife->buffer + mife->actbeg + delta, mife->buffer, mife->actlen - delta);
            mife->actbeg = 0;
            mife->actlen -= delta;
            mife->offset += delta;
            mife->ofpoint = mife->offset;
            if  (mife->eol) mife->eol->delta(mife, -delta);
            delta = 0;
        }
        if  ((mife->actbeg + mife->actlen + tmplen) > mife->buflen) {
            if  (!(mife->flags & MIFE_ALLC)) {
                if  (!(tmpbuf = malloc(mife->actbeg + mife->actlen + tmplen))) {
                    ifBLIN_QW0("malloc(%"BLIN_U")", mife->buflen);
                    ERROUT(-EX_OSERR, ENOMEM);
                }
                mife->flags |= MIFE_ALLC;
                bcopy(mife->buffer + mife->actbeg + delta, tmpbuf, mife->actlen - delta);
                mife->actbeg = 0;
                mife->buffer = tmpbuf;
                mife->buflen = mife->actbeg + mife->actlen + tmplen;
                if  (mife->eol) mife->eol->delta(mife, -delta);
            } else {
                mife->buflen = mife->actbeg + mife->actlen + tmplen;
                if  (!(mife->buffer = reallocf(mife->buffer, mife->buflen))) {
                    ifBLIN_QX0("realloc(%"BLIN_X", %"BLIN_U")", mife->buffer, mife->buflen);
                    ERROUT(-EX_OSERR, ENOMEM);
        }   }   }
        MIFE_DUMP(mife, delta);
        if  (0 > (rdlen = rd(mife, mife->buffer + mife->actbeg + mife->actlen, trylen))) {
            ex = rdlen;
            goto out;
        }
    ;   if  (!rdlen) {
    ;       if  (offset > mife->offset + (off_t)mife->actlen) {
    ;           ex = -EX_NOINPUT;
    ;           errno = ESPIPE;
    ;           goto out;
    ;       }
    ;       ifBLIN_QX3("needlen %"BLIN_D" tmplen %"BLIN_D" trylen %"BLIN_D, needlen, tmplen, trylen);
            if  (mife->flags & MIFE_FULL) CHARBUF[mife->actbeg + mife->actlen] = 0;
    ;       MIFE_DUMP(mife, delta);
            break;
    ;   } else {
    ;       mife->actlen += rdlen;
            if  (mife->flags & MIFE_FULL) CHARBUF[mife->actbeg + mife->actlen] = 0;
    ;       if  (mife->eol) mife->eol->start(mife, mife->actbeg + mife->actlen);
    ;       ifBLIN_QX3("needlen %"BLIN_D" tmplen %"BLIN_D" trylen %"BLIN_D, needlen, tmplen, trylen);
    ;       if  (mife->eol && (0 <= mife->eol->detect(mife))) {
    ;           ifBLIN_QX5(" eolp=%"BLIN_D", eolst=%"BLIN_D, mife->eol->p, mife->eol->st);
    ;           break;
    ;       }
    ;       if  (ordlen && (((off_t)mife->actlen - delta) >= ordlen)) break;
    ;       MIFE_DUMP(mife, delta);
    }   }
    ex = mife->actlen - delta;
out:
    ifBLIN_QX2( "- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static ssize_t
mife_mmap(mife_descriptor *mife, ssize_t ordlen, off_t offset) {
#   define blin_internal_flags (mife->flags)
    ssize_t  needlen;
    ssize_t  ex = -1;
    off_t    offtop;
    size_t   psz;

    ifBLIN_QX2("+ %"BLIN_X" ordlen=%"BLIN_D" offset=%"BLIN_O"d", BLIN_I(mife), ordlen, offset);
    MIFE_DUMP(mife, (off_t)0);
    if  (offset >= mife->st_size) {
        ifBLIN_QX1("offset %lld >= st_size %lld", offset, mife->st_size);
        mife->flags |= MIFE_EOFL;
        if  ((mife->flags & MIFE_FULL) && mife->buffer) CHARBUF[mife->actbeg + mife->actlen] = 0;
        ERROUT(-EX_NOINPUT, ESPIPE);
    }
    psz = getpagesize() - 1;
    if  (mife->buffer) {
        ifBLIN_QX2("  -> munmap %"BLIN_X"[%"BLIN_U"]", mife->buffer, mife->buflen);
        if  (0 > munmap(mife->buffer, mife->buflen)) {
            ifBLIN_QW0("munmap");
            EXOUT(-EX_IOERR);
        }
        mife->buffer = NULL;
    }
    mife->offset = offset & ~psz;
    mife->ofpoint = offset; /* mife->thepoint = offset - mife->offset;                          XXXX */
    mife->actbeg = 0;
    if  (!ordlen) {
        if  (mife->flags & MIFE_SLID) {
            needlen = mife->mindelta;
        } else {
            needlen = mife->st_size - offset;
            mife->flags |= MIFE_EOFL;
        }
    } else if (ordlen < mife->mindelta) {
        needlen = mife->mindelta;
    } else {
        needlen = ordlen;
    }
    ifBLIN_QX5(" needlen %"BLIN_U, needlen);
    for (;;) {
        void       *tmpbuf;

    ;   if  (offset + needlen >= mife->st_size) {
    ;       mife->actlen = mife->st_size - mife->offset;
    ;       offtop = (mife->st_size + ((mife->flags & MIFE_FULL) ? 1 : 0) + psz) & ~psz;
    ;       mife->flags |= MIFE_EOFL;
    ;   } else {
    ;       offtop = (offset + needlen + psz) & ~psz;
    ;       mife->actlen = offtop - mife->offset;
    ;       if  (mife->flags & MIFE_FULL) offtop += psz + 1;
    ;   }
    ;   mife->buflen = offtop - mife->offset;
    ;   if  (!(~mife->flags & (MIFE_EOFL | MIFE_FULL)) && !(mife->buflen & psz)) {
    ;       mife->buflen += psz + 1;
    ;       mife->buflen &= ~psz;
    ;       if  (!(mife->buffer = aligned_alloc(psz + 1, mife->buflen))) {
    ;           ifBLIN_QW0("aligned_alloc(%"BLIN_U", %"BLIN_U")", psz + 1, mife->buflen);
    ;           EXOUT(-EX_OSERR);
    ;       }
    ;       ifBLIN_QX2( "  -> alloc over %"BLIN_X"[%"BLIN_U"]", mife->buffer, mife->buflen);
    ;       mife->actbeg = mife->buflen & psz;
    ;       tmpbuf = mmap( mife->buffer + mife->actbeg
                         , mife->actlen
                         , PROT_READ | PROT_WRITE
                         , MAP_PRIVATE | MAP_FIXED
                         , mife->f.fd
                         , mife->offset
                         )
            ;
    ;       ifBLIN_QX2( "  -> mmap over tmpbuf=%"BLIN_X, tmpbuf);
    ;       MIFE_DUMP(mife, (off_t)0);
    ;       if  (tmpbuf == MAP_FAILED) {
    ;           EXOUT(-EX_IOERR);
    ;       }
    ;       mife->buffer = tmpbuf;
    ;   } else {
    ;       tmpbuf = mmap( NULL
                         , mife->buflen
                         , PROT_READ | PROT_WRITE
                         , MAP_PRIVATE
                         , mife->f.fd
                         , mife->offset
                         )
            ;
    ;       ifBLIN_QX2( "  -> mmap %d @ %"BLIN_O"d %"BLIN_X">>%"BLIN_X"[%"BLIN_U"]"
                      , mife->f.fd
                      , mife->offset
                      , mife->buffer
                      , tmpbuf
                      , mife->buflen
                      )
            ;
    ;       if  (tmpbuf == MAP_FAILED) {
    ;           ifBLIN_QW0( "mmap(NULL, %"BLIN_X"(%"BLIN_U"), %X, %X, %d, %"BLIN_O"d)"
                          , mife->buflen
                          , mife->buflen
                          , PROT_READ | PROT_WRITE
                          , MAP_PRIVATE
                          , mife->f.fd
                          , mife->offset
                          )
                ;
    ;           EXOUT(-EX_IOERR);
    ;       }
    ;       mife->buffer = tmpbuf;
    ;   }
    ;   if  (!(mife->flags & MIFE_EOFL) && !ordlen && mife->eol) {
    ;       /* mife->eol->start()  ,    */
    ;       mife->eol->start(mife, mife->ofpoint - mife->offset);             /* mife->thepoint XXXX */
    ;       if  (0 > mife->eol->detect(mife)) {
    ;           needlen *= 2;
    ;           ifBLIN_QX2( "  -> munmap %"BLIN_X"[%"BLIN_U"] needlen "BLIN_U
                          , mife->buffer
                          , mife->buflen
                          , needlen
                          )
                ;
    ;           if  (0 > munmap(mife->buffer, mife->buflen)) {
    ;               ifBLIN_QW2("munmap");
    ;               EXOUT(-EX_IOERR);
    ;           }
    ;           continue;
    ;   }   }
    ;   ex = mife->offset + mife->actlen - mife->ofpoint; /* mife->actlen - mife->thepoint;     XXXX */
    ;   if  (mife->flags & MIFE_FULL) CHARBUF[mife->actbeg + mife->actlen] = 0;
    ;   break;
    }
out:
    ifBLIN_QX2( "- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

#   define blin_internal_flags ((mife ? mife->flags : mife_default) & BLIN_MASK)

ssize_t
mife_read(mife_descriptor *mife, ssize_t ordlen, off_t offset) {
    ssize_t      ex = -1;
    struct stat  sb;
    int          i;

    ifBLIN_QX2( "+ %"BLIN_X" ordlen=%"BLIN_D" offset=%"BLIN_O"d page=%d"
              , BLIN_I(mife)
              , ordlen
              , offset
              , getpagesize()
              )
    ;
    if  (!mife) {
        ifBLIN_QX0("NULL mife descriptor");
        ERROUT(-EX_USAGE, EFAULT);
    }
    MIFE_DUMP(mife, (off_t)0);
    if  (0 > ordlen) {
        ifBLIN_QX0("0 > ordlen%"BLIN_D, ordlen);
        ERROUT(-EX_USAGE, EINVAL);
    }
    if  (  !!mife->buffer
        && (offset >= mife->offset)
        && (offset <= mife->offset + (off_t)mife->actlen)
        && (  (!!ordlen && (offset + ordlen <= mife->offset + (off_t)mife->actlen))
           || (!!mife->eol && (0 <= mife->eol->detect(mife)))
           || (mife->flags & MIFE_EOFL)
        )  ) {
        ifBLIN_QX2("  -> none");
        mife->ofpoint = offset; /* mife->thepoint = offset - mife->offset;                      XXXX */
        ex = mife->offset + mife->actlen - offset; /* mife->actlen - mife->thepoint;            XXXX */
        goto out;
    }
    if  (mife->flags & (MIFE_EOFL | MIFE_INMM)) {
        ifBLIN_QX2("  -> eof");
        ex = -EX_USAGE;
        if  (!mife->buffer) {
            ifBLIN_QX0("NULL buffer");
            errno = EDOOFUS;
        } else if (mife->offset > offset) {
            ifBLIN_QX0("read backward");
            errno = ESPIPE;
        } else if (offset > mife->offset + (off_t)mife->actlen) {
            ifBLIN_QX1("read after end");
            mife->actbeg += mife->actlen;
            mife->offset += mife->actlen;
            mife->ofpoint = mife->offset;
            mife->actlen = 0;
            errno = EDOM;
            ex = -1;
        } else {
            ifBLIN_QX0("EOFL | INMM");
            errno = EINVAL;
        }
        goto out;
    }
    if  (!(mife->flags & (MIFE_PIPE | MIFE_MMAP))) {
        if  (fstat(mife->f.fd, &sb)) {
            ifBLIN_QW0("fstat");
            EXOUT(-EX_IOERR);
        }
        mife->st_mode = sb.st_mode;
        switch (mife->st_mode & S_IFMT) {
        case S_IFREG: case S_IFDIR:
            ifBLIN_QX4("  -> mmap %"BLIN_O"d mode %07o", sb.st_size, sb.st_mode);
            mife->flags |= MIFE_MMAP;
            mife->st_size = sb.st_size;
            break;
        case S_IFIFO: case S_IFCHR: case S_IFBLK: case S_IFSOCK:
            ifBLIN_QX4("  -> pipe");
            mife->flags |= MIFE_PIPE;
            break;
        default: case S_IFLNK: case S_IFWHT: {
            ifBLIN_QX0("  -> unknown %07o", sb.st_mode);
            ERROUT(-EX_OSERR, EINVAL);
    }   }   }
    if  (mife->flags & MIFE_MMAP) {
        ex = mife_mmap(mife, ordlen, offset);
    } else if (mife->flags & MIFE_PIPE) {
        ex = mife_pipe(mife, ordlen, offset);
    }
out:
    if  (0 > ex) {
        ifBLIN_QW0("Err");
    } else if (0 > (i = mife_ctltest(mife))) {
        ifBLIN_QW0("Illegal mife");
        ex = i;
    } else if (!!i) {
        ifBLIN_QX0("Incorrect mife");
        errno = EINVAL;
        ex = -i;
    }
    ifBLIN_QX2( "- %d", ex);
    return(ex);
}

void *
mife_pointer(mife_descriptor *mife) {
    void  *p = NULL;

    ifBLIN_QX2("+ %"BLIN_X, BLIN_I(mife));
    if  (!!mife_ctltest(mife)) {
        ifBLIN_QW0("Illegal mife");
        goto out;
    }
    MIFE_DUMP(mife, (off_t)0);
    p = mife->buffer + mife->ofpoint - mife->offset + mife->actbeg;
out:
    ifBLIN_QX2("- %"BLIN_X, BLIN_I(p));
    return(p);
}

void *
mife_get(mife_descriptor *mife, off_t offset) {
    void  *p = NULL;

    ifBLIN_QX2("+ %"BLIN_X" %"BLIN_O"d", BLIN_I(mife), offset);
    if  (!!mife_ctltest(mife)) {
        ifBLIN_QW0("Illegal mife");
        goto out;
    }
    MIFE_DUMP(mife, (off_t)0);
    if  (!!mife_ctlofft(mife, offset)) {
        ifBLIN_QW0("ill offset=%"BLIN_O"d", offset);
        goto out;
    }
    mife->ofpoint = offset;
    p = mife->buffer + offset - mife->offset + mife->actbeg;
out:
    ifBLIN_QX2("- %"BLIN_X, BLIN_I(p));
    return(p);
}

void *
mife_window(mife_descriptor *mife, off_t offset, ssize_t ordlen) {
    ssize_t  tmplen;
    void    *p = NULL;

    ifBLIN_QX2("+ %"BLIN_X" %"BLIN_O"d[%"BLIN_U"]", BLIN_I(mife), offset, ordlen);
    MIFE_DUMP(mife, (off_t)0);
    if  (0 > (tmplen = mife_read(mife, ordlen, offset))) {
        ifBLIN_QW0("mife_read(,%"BLIN_D",%"BLIN_O"d)", ordlen, offset);
        goto out;
    }
    if  (!!ordlen && (tmplen < ordlen)) {
        ifBLIN_QX1("short file %"BLIN_D" < %"BLIN_D, tmplen, ordlen);
    }
    p = mife->buffer + offset - mife->offset + mife->actbeg;
out:
    ifBLIN_QX2("- %"BLIN_X, BLIN_I(p));
    return(p);
}

off_t
mife_offset(mife_descriptor *mife) {
    return(mife->ofpoint);
}
